from pathlib import Path
from typing import List

from setuptools import find_packages, setup

BASE = Path(__file__).parent
REQUIREMENTS_DIR = BASE / 'requirements'
LONG_DESCRIPTION = (BASE / 'README.md').read_text()


def get_requirements(filename: str) -> List[str]:
    with open(REQUIREMENTS_DIR / filename) as file:
        return [ln.strip() for ln in file.readlines()]


essential_packages = get_requirements('essential.txt')
api_service_packages = get_requirements('docker/api.txt')
dashboard_service_packages = get_requirements('docker/dashboard.txt')
standalone_service_packages = get_requirements('docker/standalone.txt')
worker_service_packages = get_requirements('docker/worker.txt')

demo_packages = get_requirements('demo.txt')
dev_packages = get_requirements('dev.txt')
docs_packages = get_requirements('docs.txt')
train_packages = get_requirements('train.txt')

setup(
    name='trecover',
    version='1.0.1',
    license='Apache License 2.0',
    author='Alexander Shulga',
    author_email='alexandershulga.sh@gmail.com',
    url='https://github.com/alex-snd/TRecover',
    description='A python library for training a Transformer neural network to solve the'
                ' Running Key Cipher, widely known in the field of cryptography.',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    keywords=[
        'Deep Learning',
        'Machine Learning',
        'Transformers',
        'NLP',
        'Cryptography',
        'Keyless Reading',
        'TRecover',
        'Text Recovery',
        'PyTorch',
    ],
    python_requires='>=3.8',
    packages=find_packages(
        where='src'
    ),
    package_dir={'': 'src'},
    install_requires=[essential_packages],
    extras_require={
        'api': api_service_packages,
        'dashboard': dashboard_service_packages,
        'standalone': standalone_service_packages,
        'worker': worker_service_packages,
        "demo": demo_packages,
        "dev": dev_packages,
        "docs": docs_packages,
        "train": train_packages,
    },
    entry_points={
        'console_scripts': [
            'trecover = trecover.app.cli.trecovercli:cli',
        ],
    },
    classifiers=[
        'Development Status :: 4 - Beta',

        'Environment :: Console',
        'Environment :: GPU :: NVIDIA CUDA',
        'Environment :: Web Environment',

        'Framework :: AsyncIO',
        'Framework :: Celery',
        'Framework :: FastAPI',
        'Framework :: Jupyter',

        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',

        'License :: OSI Approved :: Apache Software License',

        'Natural Language :: English',

        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Unix',

        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',

        'Topic :: Education',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Security',
        'Topic :: Security :: Cryptography',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



from pathlib import Path

from typer import Typer, Option

from trecover.config import var, log

cli = Typer(name='Download-cli', add_completion=False, help='Download train data or pre-trained model')


@cli.command(name='data', help='Download train data')
def download_data(link: str = Option(var.TRAIN_DATA_URL, help='Link to the train data on Yandex disk or GitHub'),
                  save_dir: Path = Option(var.DATA_DIR, help='Path where to store downloaded data'),
                  yandex_disk: bool = Option(False, is_flag=True, help='If the link is to Yandex disk')
                  ) -> None:
    """
    Download train data from Yandex disk or GitHub.

    Parameters
    ----------
    link : str, default=var.TRAIN_DATA_URL
        Sharing link to the train data on Yandex disk or GitHub.
    save_dir : Path, default=var.DATA_DIR
        Path where to store downloaded data.
    yandex_disk : bool, default=False
        If the link is to Yandex disk.

    """

    from trecover.utils.cli import download_archive

    download_archive(link=link, save_dir=save_dir, yandex_disk=yandex_disk)


@cli.command(name='artifacts', help='Download model artifacts by specified version or archive_link')
def download_artifacts(version: str = Option('latest', help="Artifacts' version"),
                       archive_link: str = Option(None, help='Link to the artifacts archive on Yandex disk or GitHub'),
                       save_dir: Path = Option(var.INFERENCE_DIR, help='Path where to save downloaded artifacts'),
                       yandex_disk: bool = Option(False, is_flag=True, help='If the archive_link is to Yandex disk'),
                       show: bool = Option(False, is_flag=True, help="Print available artifacts' versions")
                       ) -> None:
    """
    Download model artifacts by specified version or archive_link to Yandex disk or GitHub.

    Parameters
    ----------
    version : str, default='latest'
        Artifacts' version.
    archive_link : str, default=None
        Sharing link to the model artifacts archive on Yandex disk or GitHub.
    save_dir : Path, default=var.INFERENCE_DIR
        Path where to save downloaded artifacts.
    yandex_disk : bool, default=False
        If the link is to Yandex disk.
    show : bool, default=False
        Print available artifacts' versions.

    """

    from rich.prompt import Confirm
    from trecover.utils.cli import download_archive, download_from_github

    if show:
        log.project_console.print(var.CHECKPOINT_URLS.keys())

    elif archive_link:
        download_archive(link=archive_link, save_dir=save_dir, yandex_disk=yandex_disk)

    elif version in var.CHECKPOINT_URLS:
        download_from_github(direct_link=var.CHECKPOINT_URLS[version]['model'], save_dir=save_dir)
        download_from_github(direct_link=var.CHECKPOINT_URLS[version]['config'], save_dir=save_dir)

    elif Confirm.ask(prompt='[bright_blue]Specified version was not found. Continue downloading the latest version?',
                     default=True,
                     console=log.project_console):
        download_from_github(direct_link=var.CHECKPOINT_URLS['latest']['model'], save_dir=save_dir)
        download_from_github(direct_link=var.CHECKPOINT_URLS['latest']['config'], save_dir=save_dir)


if __name__ == '__main__':
    cli()



import re
from argparse import Namespace
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Tuple, Union, Optional

import numpy as np
import torch
from torch import Tensor
from torch.optim import Optimizer


class ExperimentParams(dict):
    """ Container for experiment parameters """

    def __init__(self, params: Union[Namespace, Dict, None] = None):
        super(ExperimentParams, self).__init__()

        if isinstance(params, Namespace):
            self.__dict__.update(vars(params))
        if isinstance(params, dict):
            self.__dict__.update(params)

    def __getitem__(self, key: Any) -> Any:
        return self.__dict__[key]

    def __repr__(self):
        return repr(self.__dict__)

    def __str__(self):
        return str(self.__dict__)

    def jsonify(self) -> Dict:
        """
        Simplify experiment parameters for further json serialization.

        Returns
        -------
            Experiment parameters as a dictionary with python built-in types values.

        """
        simplified = dict()
        simplified.update(self.__dict__)

        for key, value in self.__dict__.items():
            if isinstance(value, Path):
                simplified[key] = str(value)

        return simplified

    def update(self, *args: Any, **kwargs: Any) -> None:
        self.__dict__.update(*args, **kwargs)


def set_seeds(seed: int = 2531) -> None:
    """ Set seeds for experiment reproducibility """

    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)  # multi-GPU


def get_experiment_mark() -> str:
    """ Generate a string mark based on current date """

    date = datetime.now()
    return f'{date.month:0>2}-{date.day:0>2}-{date.hour:0>2}-{date.minute:0>2}'


def optimizer_to_str(optimizer: Optimizer) -> str:
    """
    Get optimizer object string representation.

    Parameters
    ----------
    optimizer : Optimizer
        Optimizer object for representation.

    Returns
    -------
    str:
        Optimizer string representation.

    """

    optimizer_str_repr = str(optimizer)

    optimizer_name = optimizer_str_repr.split()[0]
    optimizer_params = re.findall(r'(.+): (.+)', optimizer_str_repr)

    return f"{optimizer_name}({', '.join([f'{param.strip()}={value.strip()}' for param, value in optimizer_params])})"


def transfer(tensors: Tuple[Optional[Tensor], ...], to_device: torch.device) -> Tuple[Optional[Tensor], ...]:
    """
    Transfer the tensors to a specified device.

    Parameters
    ----------
    tensors : Tuple
        Sequence of tensors to transfer.
    to_device : torch.device
        The desired device of returned tensors.

    Returns
    -------
    Tuple[Optional[Tensor], ...]:
        Tuple of tensors allocated on the specified device.

    """
    return tuple([
        tensor.to(to_device) if isinstance(tensor, Tensor) else tensor
        for tensor in tensors
    ])



from abc import ABC, abstractmethod

from torch.optim import Optimizer


class BaseScheduler(ABC):
    # TODO docs

    def __repr__(self) -> str:
        return self.__str__()

    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractmethod
    def step(self) -> None:
        # TODO docs

        pass

    @abstractmethod
    def set_rate(self, rate: float) -> None:
        # TODO docs

        pass


class IdentityScheduler(BaseScheduler):
    # TODO docs

    def __str__(self) -> str:
        return '<IdentityScheduler()>'

    def step(self) -> None:
        pass

    def set_rate(self, rate: float) -> None:
        pass


class WarmupScheduler(BaseScheduler):
    # TODO docs

    def __init__(self,
                 optimizer: Optimizer,
                 d_model: int,
                 warmup: int,
                 step_size: int,
                 seek: int = 0,
                 factor: float = 1.0):
        self.optimizer = optimizer
        self.d_model = d_model
        self.warmup = warmup
        self.step_size = step_size
        self.factor = factor

        self.__step = self.__get_rate_step(self.optimizer.param_groups[0]['lr']) + seek
        self.__update_rate()

    def __str__(self) -> str:
        return f'WarmupScheduler(warmup={self.warmup}, step_size={self.step_size}, factor={self.factor})'

    def step(self) -> None:
        self.__step += 1

        if self.__step % self.step_size == 0:
            self.__update_rate()

    def set_rate(self, rate: float) -> None:
        self.__step = self.__get_rate_step(rate)
        self.__update_rate()

    def __update_rate(self) -> None:
        self.__rate = self.__get_step_rate(self.__step)

        for p in self.optimizer.param_groups:
            p['lr'] = self.__rate

    def __get_step_rate(self, step: int) -> float:
        if self.__step == 0:
            return 0

        return self.factor * (self.d_model ** (-0.5)) * min((step / self.step_size) ** (-0.5),
                                                            (step / self.step_size) * self.warmup ** (-1.5))

    def __get_rate_step(self, rate: float) -> int:
        if rate == 0:
            return 0

        return round(self.step_size / (self.d_model * rate ** 2))



import logging
from pathlib import Path

from rich.console import Console
from rich.logging import RichHandler

from trecover.config import var

var.LOGS_DIR.mkdir(parents=True, exist_ok=True)

MLFLOW_LOG = var.LOGS_DIR / 'mlflow.log'
DASHBOARD_LOG = var.LOGS_DIR / 'dashboard.log'
API_LOG = var.LOGS_DIR / 'api.log'
WORKER_LOG = var.LOGS_DIR / 'worker.log'

project_logger = logging.getLogger('project')
project_logger.setLevel(logging.DEBUG)

project_console = Console(force_terminal=True, record=True)
console_handler = RichHandler(console=project_console, markup=True, show_time=False, show_level=False, show_path=False)
console_handler.setLevel(logging.DEBUG)
project_logger.addHandler(hdlr=console_handler)

error_console = Console(file=Path(var.LOGS_DIR, 'error.log').open(mode='a'))
error_handler = RichHandler(console=error_console, markup=True, rich_tracebacks=True, tracebacks_show_locals=True)
error_handler.setLevel(logging.ERROR)
project_logger.addHandler(hdlr=error_handler)

info_console = Console(file=Path(var.LOGS_DIR, 'info.log').open(mode='a'))
info_handler = RichHandler(console=info_console, markup=True, rich_tracebacks=True, tracebacks_show_locals=True)
info_handler.setLevel(logging.INFO)
project_logger.addHandler(hdlr=info_handler)



import os
from enum import Enum
from pathlib import Path

from dotenv import load_dotenv

# -------------------------------------------------General Variables----------------------------------------------------

ALPHABET = {'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'v',
            'u', 'w', 'x', 'y', 'z'}

ALPHABET2NUM = {'a': 0, 'b': 1, 'c': 2, 'd': 3, 'e': 4, 'f': 5, 'g': 6, 'h': 7, 'i': 8, 'j': 9, 'k': 10,
                'l': 11, 'm': 12, 'n': 13, 'o': 14, 'p': 15, 'q': 16, 'r': 17, 's': 18, 't': 19, 'v': 20,
                'u': 21, 'w': 22, 'x': 23, 'y': 24, 'z': 25}

NUM2ALPHABET = {0: 'a', 1: 'b', 2: 'c', 3: 'd', 4: 'e', 5: 'f', 6: 'g', 7: 'h', 8: 'i', 9: 'j', 10: 'k',
                11: 'l', 12: 'm', 13: 'n', 14: 'o', 15: 'p', 16: 'q', 17: 'r', 18: 's', 19: 't', 20: 'v',
                21: 'u', 22: 'w', 23: 'x', 24: 'y', 25: 'z'}

COLORS = ('green', 'yellow', 'blue', 'magenta', 'cyan', 'bright_green', 'bright_yellow', 'bright_blue',
          'bright_magenta', 'bright_cyan', 'red')

CHECKPOINT_URLS = {
    'latest': {'model': 'https://github.com/alex-snd/TRecover/releases/download/v0.1.0/model.pt',
               'config': 'https://github.com/alex-snd/TRecover/releases/download/v0.1.0/config.json'},
    'v0.1.0': {'model': 'https://github.com/alex-snd/TRecover/releases/download/v0.1.0/model.pt',
               'config': 'https://github.com/alex-snd/TRecover/releases/download/v0.1.0/config.json'},
}

TRAIN_DATA_URL = 'https://github.com/alex-snd/TRecover/releases/download/v0.1.0/data.zip'

CONFIG_DIR = Path(__file__).parent.absolute()
BASE_INIT = CONFIG_DIR / 'base.init'

if not BASE_INIT.exists():
    BASE_DIR = Path().absolute()
else:
    with BASE_INIT.open() as f:
        BASE_DIR = Path(f.read()).absolute()

LOGS_DIR = BASE_DIR / 'logs'
DATA_DIR = BASE_DIR / 'data'
EXPERIMENTS_DIR = BASE_DIR / 'experiments'

load_dotenv(BASE_DIR / '.env')

# --------------------------------------------------Mlflow Variables----------------------------------------------------

MLFLOW_HOST = os.getenv('MLFLOW_HOST', default='localhost')
MLFLOW_PORT = int(os.getenv('MLFLOW_PORT', default=8002))
MLFLOW_WORKERS = int(os.getenv('MLFLOW_WORKERS', default=1))
MLFLOW_PID = CONFIG_DIR / 'mlflow.pid'

# -------------------------------------------------Dashboard Variables--------------------------------------------------

STREAMLIT_HOST = os.getenv('STREAMLIT_HOST', default='localhost')
STREAMLIT_PORT = int(os.getenv('STREAMLIT_PORT', default=8000))
DASHBOARD_PID = CONFIG_DIR / 'dashboard.pid'

# ---------------------------------------------------API Variables------------------------------------------------------

FASTAPI_HOST = os.getenv('FASTAPI_HOST', default='localhost')
FASTAPI_PORT = int(os.getenv('FASTAPI_PORT', default=8001))
FASTAPI_WORKERS = int(os.getenv('FASTAPI_WORKERS', default=1))
FASTAPI_URL = f'http://{FASTAPI_HOST}:{FASTAPI_PORT}'
API_PID = CONFIG_DIR / 'api.pid'

# --------------------------------------------------Broker Variables----------------------------------------------------

BROKER_PORT = int(os.getenv('BROKER_PORT', default=5672))
BROKER_UI_PORT = int(os.getenv('BROKER_UI_PORT', default=15672))
BROKER_IMAGE = os.getenv('BROKER_IMAGE', default='rabbitmq:3.9.8-management')
BROKER_VOLUME_ID = 'trecover_broker_data'
BROKER_ID = 'trecover_broker'

# -------------------------------------------------Backend Variables----------------------------------------------------

BACKEND_PORT = int(os.getenv('BACKEND_PORT', default=6379))
BACKEND_IMAGE = os.getenv('BACKEND_IMAGE', default='redis:6.2')
BACKEND_VOLUME_ID = 'trecover_backend_data'
BACKEND_ID = 'trecover_backend'

# --------------------------------------------------Worker Variables----------------------------------------------------

CELERY_BROKER = os.getenv('CELERY_BROKER', default='pyamqp://guest@localhost')
CELERY_BACKEND = os.getenv('CELERY_BACKEND', default='redis://localhost')
CELERY_WORKERS = int(os.getenv('CELERY_WORKERS', default=1))
WORKER_PID = CONFIG_DIR / 'worker.pid'

# -------------------------------------------------Inference Variables--------------------------------------------------

INFERENCE_DIR = Path(os.getenv('INFERENCE_DIR', default=BASE_DIR / 'inference'))
INFERENCE_PARAMS_PATH = Path(os.getenv('INFERENCE_PARAMS_PATH', default=INFERENCE_DIR / 'config.json'))
INFERENCE_WEIGHTS_PATH = Path(os.getenv('INFERENCE_WEIGHTS_PATH', default=INFERENCE_DIR / 'model.pt'))
CUDA = os.getenv('CUDA', default='').lower() != 'false'
MAX_NOISE = int(os.getenv('MAX_NOISE', default=13))


# -------------------------------------------------Services Variables---------------------------------------------------

class PoolType(str, Enum):
    prefork = 'prefork'
    eventlet = 'eventlet'
    gevent = 'gevent'
    processes = 'processes'
    solo = 'solo'


class LogLevel(str, Enum):
    debug = 'debug'
    info = 'info'
    warning = 'warning'
    error = 'error'
    critical = 'critical'


DEFAULT_CONFIG = {
    'dashboard': {'host': STREAMLIT_HOST, 'port': STREAMLIT_PORT, 'loglevel': 'info'},
    'api': {'host': FASTAPI_HOST, 'port': FASTAPI_PORT, 'loglevel': 'info', 'concurrency': FASTAPI_WORKERS},
    'worker': {'name': 'TRecoverWorker', 'pool': 'solo', 'loglevel': 'info', 'concurrency': CELERY_WORKERS,
               'broker_url': CELERY_BROKER, 'backend_url': CELERY_BACKEND},
    'broker': {'port': BROKER_PORT, 'ui_port': BROKER_UI_PORT, 'auto_remove': False},
    'backend': {'port': BACKEND_PORT, 'auto_remove': False}
}



from typing import List, Union


def visualize_columns(columns: List[str], delimiter: str = '', as_rows=False) -> Union[str, List[str]]:
    """
    Get the columns string representation.

    Parameters
    ----------
    columns : List[str]
        Columns for keyless reading.
    delimiter : str
        Delimiter for columns visualization.
    as_rows : bool
        Return visualization as a list of strings.

    Returns
    -------
    Union[str, List[str]]:
        Visualization as a string or list of strings.

    """

    if len(columns) == 0:
        return ''

    rows = list()
    columns = [list(column) for column in columns]
    max_depth = max([len(column) for column in columns])

    for d in range(max_depth):

        row = str()
        for c in range(len(columns)):
            row += f'{delimiter}{columns[c][d]}' if d < len(columns[c]) else f'{delimiter} '

        rows.append(f'{row}{delimiter}')

    return rows if as_rows else '\n'.join(rows)


def visualize_target(target: List[str], delimiter: str = '') -> str:
    """
    Get the target string representation.

    Parameters
    ----------
    target : List[str]
        List of correct symbols for each column.
    delimiter : str
        Delimiter for target visualization.

    Returns
    -------
    str:
        Visualization as a string.

    """

    return f'{delimiter}{delimiter.join(target)}{delimiter}'



import shutil
import tempfile
import traceback
from pathlib import Path
from time import time
from typing import Callable, Optional, Tuple, Union, Type

import torch
from rich.console import Group, Console
from rich.panel import Panel
from rich.progress import Progress, TextColumn, BarColumn, TimeElapsedColumn, TimeRemainingColumn
from rich.text import Text
from torch import Tensor
from torch.optim import Optimizer
from torch.utils.data import DataLoader

from trecover.config import log
from trecover.model import TRecover
from trecover.train.monitor import BaseMonitor, IdentityMonitor
from trecover.train.scheduler import BaseScheduler, IdentityScheduler
from trecover.utils.model import save_params
from trecover.utils.train import ExperimentParams, optimizer_to_str, transfer
from trecover.utils.transform import tensor_to_columns, tensor_to_target
from trecover.utils.visualization import visualize_columns, visualize_target


# TODO docs after RemoteTrainer implementation using pytorch-lightning


class LocalTrainer(object):
    def __init__(self,
                 params: ExperimentParams,
                 model: TRecover,
                 criterion: Callable[..., Tensor],
                 optimizer: Optimizer,
                 exp_dir: Path,
                 scheduler: Optional[BaseScheduler] = None,
                 monitor: Optional[BaseMonitor] = None,
                 device: Optional[torch.device] = None,
                 accumulation_step: int = 1,
                 log_interval: int = 1,
                 saving_interval: int = 1,
                 vis_interval: int = 1,
                 n_columns_to_show: Optional[int] = None,
                 delimiter: str = '',
                 console: Console = log.project_console):
        self.params = params
        self.model = model
        self.criterion = criterion  # should return average on the batch
        self.optimizer = optimizer
        self.exp_dir = exp_dir
        self.scheduler = scheduler or IdentityScheduler()
        self.monitor = monitor or IdentityMonitor()
        self.device = device or torch.device("cpu")
        self.accumulation_step = accumulation_step
        self.log_interval = log_interval
        self.saving_interval = saving_interval
        self.vis_interval = vis_interval
        self.n_columns_to_show = n_columns_to_show
        self.delimiter = delimiter

        self.experiment_folder = self.exp_dir / self.monitor.experiment_name
        self.weights_folder = self.experiment_folder / 'weights'

        self.experiment_folder.mkdir(parents=True, exist_ok=True)
        self.weights_folder.mkdir(parents=True, exist_ok=True)

        self.log_file = self.experiment_folder / f'log_{self.monitor.experiment_name}.html'
        self.console = console

        self.__log_init_params()

    def __enter__(self) -> Tuple['LocalTrainer', BaseMonitor]:
        self.monitor.start()

        return self, self.monitor

    def __exit__(self, exc_type: Union[None, Type[BaseException]],
                 exc_val: Union[None, BaseException],
                 exc_tb: traceback.TracebackException
                 ) -> None:
        self.save_html_log()

        with tempfile.TemporaryDirectory() as dp:
            save_params(self.params.jsonify(), Path(dp, 'params.json'))
            self.model.save(Path(dp, 'model.pt'))
            shutil.copy(self.log_file, dp)

            self.monitor.log_artifact(dp)

        self.monitor.finish()

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return f'Trainer(model={self.model}, ' \
               f'criterion={self.criterion}, ' \
               f'optimizer={optimizer_to_str(self.optimizer)}), ' \
               f'exp_dir={self.exp_dir}, ' \
               f'scheduler={self.scheduler}, ' \
               f'device={self.device}, ' \
               f'accumulation_step={self.accumulation_step}, ' \
               f'log_interval={self.log_interval}, ' \
               f'saving_interval={self.saving_interval}, ' \
               f'vis_interval={self.vis_interval}, ' \
               f'n_columns_to_show={self.n_columns_to_show}, ' \
               f'delimiter={self.delimiter})'

    @property
    def __lr(self) -> float:
        return self.optimizer.param_groups[0]['lr']

    def __log_init_params(self) -> None:
        self.console.print(f'Date: {self.monitor.experiment_name}')
        self.console.print(f'Model: {self.model}')
        self.console.print(f'Trainable parameters: {self.model.params_count:,}')
        self.console.print(f'Optimizer: {optimizer_to_str(self.optimizer)}')
        self.console.print(f'Accumulation step: {self.accumulation_step}')
        self.console.print(f'Scheduler: {self.scheduler}')

    def __train_step(self, offset: int, train_loader: DataLoader, accumulation_step: int = 1) -> None:
        self.model.train()

        self.console.rule(Text('Training step', style='bold magenta'), style='magenta')

        train_loss = 0.0

        self.optimizer.zero_grad(set_to_none=True)
        start_time = time()

        for batch_idx, train_tensors in enumerate(train_loader, start=1):
            src, tgt_inp, tgt, src_pad_mask, tgt_pad_mask, tgt_attn_mask = transfer(train_tensors,
                                                                                    to_device=self.device)

            tgt_out = self.model(src, src_pad_mask, tgt_inp, tgt_attn_mask, tgt_pad_mask)
            tgt_out = tgt_out.reshape(-1, self.model.token_size)
            tgt = tgt.view(-1)
            src = src.reshape(-1, self.model.token_size)

            loss = self.criterion(src, tgt_inp, tgt, src_pad_mask, tgt_pad_mask, tgt_attn_mask, tgt_out)
            loss.backward()

            train_loss += loss.item()

            if batch_idx % accumulation_step == 0:
                self.scheduler.step()
                self.optimizer.step()

                self.optimizer.zero_grad(set_to_none=True)

                if (offset + batch_idx) % self.log_interval == 0:
                    accuracy = (torch.argmax(tgt_out, dim=1) == tgt).float().sum().item() / tgt.size(0)
                    train_loss /= accumulation_step

                    self.console.print(f'Train Batch:  {offset + batch_idx:^7} | '
                                       f'Loss: {train_loss:>10.6f} | Accuracy: {accuracy:>6.3f} | '
                                       f'Elapsed: {time() - start_time:>7.3f} | LR {round(self.__lr, 6):>8}')

                    self.monitor.log_metrics({'Train loss': train_loss,
                                              'Train accuracy': accuracy,
                                              'LR': self.__lr,
                                              'train_step': offset + batch_idx},
                                             step=offset + batch_idx)

                    train_loss = 0.0
                    start_time = time()

    @torch.no_grad()
    def __val_step(self, offset: int, val_loader: DataLoader) -> None:
        self.model.eval()

        self.console.rule(Text('Validation step', style='bold magenta'), style='magenta')

        val_loss = 0
        val_accuracy = 0
        start_time = time()

        for batch_idx, val_tensors in enumerate(val_loader, start=1):
            src, tgt_inp, tgt, src_pad_mask, tgt_pad_mask, tgt_attn_mask = transfer(val_tensors, to_device=self.device)

            tgt_out = self.model(src, src_pad_mask, tgt_inp, tgt_attn_mask, tgt_pad_mask)
            tgt_out = tgt_out.reshape(-1, self.model.token_size)
            tgt = tgt.view(-1)
            src = src.reshape(-1, self.model.token_size)

            loss = self.criterion(src, tgt_inp, tgt, src_pad_mask, tgt_pad_mask, tgt_attn_mask, tgt_out).item()
            accuracy = ((torch.argmax(tgt_out, dim=1) == tgt).float().sum() / tgt.size(0)).item()

            val_loss += loss
            val_accuracy += accuracy

            self.monitor.log_metrics({"Val loss": loss,
                                      "Val accuracy": accuracy,
                                      'val_step': offset + batch_idx},
                                     step=offset + batch_idx)

            self.console.print(f'Val Batch:    {offset + batch_idx:^7} | Loss: {loss:>10.6f} | '
                               f'Accuracy: {accuracy:>6.3f} | Elapsed: {time() - start_time:>7.3f}')

            start_time = time()

        val_loss /= len(val_loader)
        val_accuracy /= len(val_loader)

        self.console.print(Panel(f'Val Loss:    {val_loss:>10.6f} \nVal Accuracy: {val_accuracy:>6.3f}',
                                 title='Validation average', highlight=True, border_style='magenta'))

    @torch.no_grad()
    def __vis_step(self, vis_loader: DataLoader) -> None:
        self.model.eval()

        self.console.rule(Text('Visualization step', style='bold magenta'), style='magenta')

        for batch_idx, vis_tensors in enumerate(vis_loader, start=1):
            src, tgt_inp, tgt, src_pad_mask, tgt_pad_mask, tgt_attn_mask = transfer(vis_tensors, to_device=self.device)

            tgt_out = self.model(src, src_pad_mask, tgt_inp, tgt_attn_mask, tgt_pad_mask)
            tgt_out = tgt_out.reshape(-1, self.model.token_size)
            prediction = torch.argmax(tgt_out, dim=1).view_as(tgt)

            for i in range(src.size(0)):
                columns = tensor_to_columns(src[i, : self.n_columns_to_show])
                columns = visualize_columns(columns, delimiter=self.delimiter, as_rows=True)
                columns = (Text(row, style='bright_blue', overflow='ellipsis', no_wrap=True) for row in columns)
                target = tensor_to_target(prediction[i, : self.n_columns_to_show])
                predicted = visualize_target(target, delimiter=self.delimiter)
                original = tensor_to_target(tgt[i, : self.n_columns_to_show])
                original = visualize_target(original, delimiter=self.delimiter)

                panel_group = Group(
                    Text('Columns', style='magenta', justify='center'),
                    *columns,
                    Text('Predicted', style='magenta', justify='center'),
                    Text(predicted, style='cyan', justify='center', overflow='ellipsis'),
                    Text('Original', style='magenta', justify='center'),
                    Text(original, justify='center', overflow='ellipsis')
                )

                self.console.print(
                    Panel(panel_group, title=f'Example {(batch_idx - 1) * vis_loader.batch_size + i + 1}',
                          border_style='magenta'),
                    justify='center'
                )
                self.console.print('\n')

    def train(self,
              n_epochs: int,
              train_loader: DataLoader,
              val_loader: Optional[DataLoader] = None,
              vis_loader: Optional[DataLoader] = None,
              epoch_seek: int = 0
              ) -> None:
        if len(train_loader) % self.accumulation_step != 0:
            self.console.print('WARNING: Train dataset size must be evenly divisible by batch_size * accumulation_step',
                               style='bold red')
        if self.device.type == 'cpu':
            self.console.print('WARNING: Training without GPU usage', style='bold red')

        try:
            for epoch_idx in range(epoch_seek + 1, epoch_seek + n_epochs + 1):
                train_offset = len(train_loader) * (epoch_idx - 1)
                val_offset = len(val_loader) * (epoch_idx - 1)

                self.__train_step(train_offset, train_loader, self.accumulation_step)

                if val_loader:
                    self.__val_step(val_offset, val_loader)

                if vis_loader and epoch_idx % self.vis_interval == 0:
                    self.__vis_step(vis_loader)

                if epoch_idx % self.saving_interval == 0:
                    self.save_html_log()
                    self.save_model(str(train_offset + len(train_loader)))

        except KeyboardInterrupt:
            self.console.print('Training is interrupted', style='yellow')

    @torch.no_grad()
    def test(self, test_loader: DataLoader) -> Tuple[float, float]:
        self.model.eval()

        test_loss = 0
        test_accuracy = 0

        with Progress(
                TextColumn('{task.description}', style='bright_blue'),
                BarColumn(complete_style='bright_blue'),
                TextColumn('{task.percentage:>3.0f}%', style='bright_blue'),
                TextColumn('Remaining', style='bright_blue'),
                TimeRemainingColumn(),
                TextColumn('Elapsed', style='bright_blue'),
                TimeElapsedColumn(),
                console=log.project_console,
                transient=True,
        ) as progress:
            test_progress = progress.add_task('Testing', total=len(test_loader))

            for batch_idx, test_tensors in enumerate(test_loader, start=1):
                src, tgt_inp, tgt, src_pad_mask, tgt_pad_mask, tgt_attn_mask = transfer(test_tensors,
                                                                                        to_device=self.device)

                tgt_out = self.model(src, src_pad_mask, tgt_inp, tgt_attn_mask, tgt_pad_mask)
                tgt_out = tgt_out.reshape(-1, self.model.token_size)
                tgt = tgt.view(-1)
                src = src.reshape(-1, self.model.token_size)

                loss = self.criterion(src, tgt_inp, tgt, src_pad_mask, tgt_pad_mask, tgt_attn_mask, tgt_out).item()
                accuracy = ((torch.argmax(tgt_out, dim=1) == tgt).float().sum() / tgt.size(0)).item()

                test_loss += loss
                test_accuracy += accuracy

                self.monitor.log_metrics({"Test loss": loss,
                                          "Test accuracy": accuracy,
                                          'test_step': batch_idx},
                                         step=batch_idx)

                progress.update(test_progress, advance=1)

        test_loss /= len(test_loader)
        test_accuracy /= len(test_loader)

        self.monitor.log_variables({'Test Loss': test_loss, 'Test Accuracy': test_accuracy})

        self.console.print(Panel(f'Test Loss:    {test_loss:>10.6f} \nTest Accuracy: {test_accuracy:>6.3f}',
                                 title='Testing', highlight=True, border_style='magenta'))

        return test_loss, test_accuracy

    def save_model(self, weights_name: str) -> None:
        self.model.save(filename=self.weights_folder / f'{self.monitor.experiment_name}_{weights_name}.pt')

    def save_html_log(self) -> None:
        self.console.save_html(str(self.log_file), clear=False)

    # TODO find_batch_size
    def find_batch_size(self):
        pass



import math
from pathlib import Path
from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor


class PositionalEncoding(nn.Module):
    def __init__(self, d_model: int, dropout: float = 0.1, max_len: int = 5000):
        super(PositionalEncoding, self).__init__()

        self.dropout = nn.Dropout(p=dropout)

        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model))
        x = position * div_term

        pe = torch.zeros(max_len, d_model)
        pe[:, 0::2] = torch.sin(x)
        pe[:, 1::2] = torch.cos(x)
        pe = pe.unsqueeze(0).transpose(0, 1)

        self.register_buffer('pe', pe)

    def forward(self, x: Tensor) -> Tensor:
        x = x + self.pe[:x.shape[0], :]

        return self.dropout(x)


class TRecover(nn.Module):
    def __init__(self,
                 token_size: int,
                 pe_max_len: int,
                 num_layers: int,
                 d_model: int,
                 n_heads: int,
                 d_ff: int,
                 dropout: float):
        assert d_model % n_heads == 0, 'd_model size must be evenly divisible by n_heads size'

        super(TRecover, self).__init__()

        self.token_size = token_size
        self.pe_max_len = pe_max_len
        self.num_layers = num_layers
        self.d_model = d_model
        self.n_heads = n_heads
        self.d_ff = d_ff
        self.dropout = dropout

        self.scale = math.sqrt(d_model)

        self.mapping = nn.Linear(in_features=token_size, out_features=d_model)

        self.pe = PositionalEncoding(d_model=d_model, dropout=dropout, max_len=pe_max_len)

        self.encoder_layer = nn.TransformerEncoderLayer(d_model=d_model, nhead=n_heads, dim_feedforward=d_ff,
                                                        dropout=dropout)
        self.encoder = nn.TransformerEncoder(encoder_layer=self.encoder_layer, num_layers=num_layers)

        self.decoder_layer = nn.TransformerDecoderLayer(d_model=d_model, nhead=n_heads, dim_feedforward=d_ff,
                                                        dropout=dropout)
        self.decoder = nn.TransformerDecoder(decoder_layer=self.decoder_layer, num_layers=num_layers)

        self.inv_mapping = nn.Linear(in_features=d_model, out_features=token_size)

        self.params_count = sum(p.numel() for p in self.parameters() if p.requires_grad)

        self.init_weights()

    def __repr__(self) -> str:
        return self.__str__()  # but standard implementation is better

    def __str__(self) -> str:
        return f'TRecover(token_size={self.token_size}, pe_max_len={self.pe_max_len}, num_layers={self.num_layers}, ' \
               f'd_model={self.d_model}, n_heads={self.n_heads}, d_ff={self.d_ff}, dropout={self.dropout})'

    def init_weights(self) -> None:
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

    def encode(self, src: Tensor, src_pad_mask: Optional[Tensor]) -> Tensor:
        src = src.transpose(0, 1)

        src = F.relu(self.mapping(src)) * self.scale
        src = self.pe(src)

        return self.encoder(src=src, src_key_padding_mask=src_pad_mask)

    def decode(self,
               tgt_inp: Tensor,
               encoded_src: Tensor,
               tgt_attn_mask: Optional[Tensor],
               tgt_pad_mask: Optional[Tensor],
               src_pad_mask: Optional[Tensor]
               ) -> Tensor:
        tgt_inp = tgt_inp.transpose(0, 1)

        tgt_inp = F.relu(self.mapping(tgt_inp)) * self.scale
        tgt_inp = self.pe(tgt_inp)

        return self.decoder(tgt=tgt_inp, memory=encoded_src, tgt_mask=tgt_attn_mask,
                            tgt_key_padding_mask=tgt_pad_mask, memory_key_padding_mask=src_pad_mask)

    def predict(self,
                tgt_inp: Tensor,
                encoded_src: Tensor,
                tgt_attn_mask: Optional[Tensor],
                tgt_pad_mask: Optional[Tensor],
                src_pad_mask: Optional[Tensor]
                ) -> Tensor:
        decoded = self.decode(tgt_inp, encoded_src, tgt_attn_mask, tgt_pad_mask, src_pad_mask)

        return self.inv_mapping(decoded).transpose(0, 1)

    def forward(self, src: Tensor,
                src_pad_mask: Optional[Tensor],
                tgt_inp: Tensor,
                tgt_attn_mask: Optional[Tensor],
                tgt_pad_mask: Optional[Tensor]
                ) -> Tensor:
        encoded_src = self.encode(src, src_pad_mask)

        return self.predict(tgt_inp, encoded_src, tgt_attn_mask, tgt_pad_mask, src_pad_mask)

    def save(self, filename: Path) -> None:
        torch.save(self.state_dict(), filename)

    def load_parameters(self, filename: Path, device: torch.device, strict: bool = True) -> None:
        self.load_state_dict(torch.load(filename, map_location=device), strict=strict)



from typer import Typer, Option, Context

from trecover.config import var

cli = Typer(name='Backend-cli', add_completion=False, help='Manage Backend service')


@cli.callback(invoke_without_command=True)
def backend_state_verification(ctx: Context) -> None:
    """
    Perform cli commands and docker engine verification (state checking).

    Parameters
    ----------
    ctx : Context
        Typer (Click like) special internal object that holds state relevant
        for the script execution at every single level.

    """

    from trecover.config import log
    from trecover.utils.docker import is_docker_running, get_container

    if not is_docker_running():
        log.project_console.print('Docker engine is not running', style='red')
        ctx.exit(1)

    elif container := get_container(var.BACKEND_ID):
        if container.status == 'running' and ctx.invoked_subcommand in ('start', None):
            log.project_console.print(':rocket: The backend service is already started', style='bright_blue')
            ctx.exit(0)

    elif ctx.invoked_subcommand is None:
        backend_start(port=var.BACKEND_PORT, auto_remove=False, attach=False)

    elif ctx.invoked_subcommand != 'start':
        log.project_console.print('The backend service is not started', style='yellow')
        ctx.exit(1)


@cli.command(name='start', help='Start service')
def backend_start(port: int = Option(var.BACKEND_PORT, '--port', '-p',
                                     help='Bind socket to this port.'),
                  auto_remove: bool = Option(False, '--rm', is_flag=True,
                                             help='Remove docker container after service exit'),
                  attach: bool = Option(False, '--attach', '-a', is_flag=True,
                                        help='Attach local standard input, output, and error streams')

                  ) -> None:
    """
    Start backend service.

    Parameters
    ----------
    port : int, default=ENV(BACKEND_PORT) or 6379
        Bind backend socket to this port.
    auto_remove : bool, default=False
        Remove backend docker container after service exit.
    attach : bool, default=False
        Attach output and error streams.

    """
    from rich.prompt import Confirm

    from trecover.config import log
    from trecover.utils.docker import get_client, get_container, get_image, pull_image

    if not (image := get_image(var.BACKEND_IMAGE)):
        with log.project_console.screen(hide_cursor=False):
            if not Confirm.ask(f"The backend image '{var.BACKEND_IMAGE}' is needed to be pulled.\nContinue?",
                               default=True):
                return

        image = pull_image(var.BACKEND_IMAGE)

    if container := get_container(var.BACKEND_ID):
        container.start()

        log.project_console.print(f'The backend service is started', style='bright_blue')

    else:
        get_client().containers.run(image=image.id,
                                    name=var.BACKEND_ID,
                                    auto_remove=auto_remove,
                                    detach=True,
                                    stdin_open=True,
                                    stdout=True,
                                    tty=True,
                                    stop_signal='SIGTERM',
                                    ports={6379: port},
                                    volumes=[f'{var.BACKEND_VOLUME_ID}:/data'])

        log.project_console.print(f'The backend service is launched', style='bright_blue')

    if attach:
        backend_attach()


@cli.command(name='stop', help='Stop service')
def backend_stop(prune: bool = Option(False, '--prune', '-p', is_flag=True,
                                      help='Prune backend.'),
                 v: bool = Option(False, '--volume', '-v', is_flag=True,
                                  help='Remove the volumes associated with the container')
                 ) -> None:
    """ Stop backend service. """

    from trecover.config import log
    from trecover.utils.docker import get_container

    get_container(var.BACKEND_ID).stop()

    log.project_console.print('The backend service is stopped', style='bright_blue')

    if prune:
        backend_prune(force=False, v=v)


@cli.command(name='prune', help='Prune docker container')
def backend_prune(force: bool = Option(False, '--force', '-f', is_flag=True,
                                       help='Force the removal of a running container'),
                  v: bool = Option(False, '--volume', '-v', is_flag=True,
                                   help='Remove the volumes associated with the container')
                  ) -> None:
    """
    Prune backend service docker container.

    Parameters
    ----------
    force : bool, default=False
        Force the removal of a running container.
    v : bool, default=False
        Remove the volumes associated with the container.

    """

    from trecover.config import log
    from trecover.utils.docker import get_container, get_volume

    container = get_container(var.BACKEND_ID)

    if container.status == 'running' and not force:
        log.project_console.print('You need to stop backend service before pruning or use --force flag', style='yellow')
    else:
        container.remove(force=force)

        if v and (volume := get_volume(var.BACKEND_VOLUME_ID)):
            volume.remove(force=force)

        log.project_console.print('The backend service is pruned', style='bright_blue')


@cli.command(name='status', help='Display service status')
def backend_status() -> None:
    """ Display backend service status. """

    from trecover.config import log
    from trecover.utils.docker import get_container

    if (container := get_container(var.BACKEND_ID)) and container.status == 'running':
        log.project_console.print(':rocket: The backend status: running', style='bright_blue')
    else:
        log.project_console.print('The backend service is not started', style='yellow')


@cli.command(name='attach', help='Attach local output stream to a service')
def backend_attach() -> None:
    """ Attach local output stream to a running backend service. """

    from trecover.config import log
    from trecover.utils.docker import get_container

    with log.project_console.screen(hide_cursor=True):
        for line in get_container(var.BACKEND_ID).attach(stream=True, logs=True):
            log.project_console.print(line.decode().strip())

    log.project_console.clear()


if __name__ == '__main__':
    cli()
